//----------------------------------------------------------------------------
//
// File:        fileio.cpp
// Date:        19-Sep-2000
// Programmer:  Marc Rousseau
//
// Description: This file contains startup code for Linux/SDL
//
// Copyright (c) 2000-2003 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//   15-Jan-2003    Moved old code to support.cpp and added new cFile classes
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include "common.hpp"
#include "logger.hpp"
#include "diskio.hpp"
#include "diskfs.hpp"
#include "fileio.hpp"

DBG_REGISTER ( __FILE__ );

static inline USHORT GetUSHORT ( const void *_ptr )
{
    FUNCTION_ENTRY ( NULL, "GetUSHORT", true );

    const UCHAR *ptr = ( const UCHAR * ) _ptr;
    return ( USHORT ) (( ptr [0] << 8 ) | ptr [1] );
}

static inline USHORT GetUSHORT_LE ( const void *_ptr )
{
    FUNCTION_ENTRY ( NULL, "GetUSHORT_LE", true );

    const UCHAR *ptr = ( const UCHAR * ) _ptr;
    return ( USHORT ) (( ptr [1] << 8 ) | ptr [0] );
}

cFile *OpenFile ( const char * )
{
    return NULL;
}


cDiskFile::cDiskFile ( cDiskFileSystem *disk, sFileDescriptorRecord *fdr ) :
    m_Disk ( disk ),
    m_FDR ( fdr ),
    m_TotalRecordsLeft ( 0 ),
    m_RecordsLeft ( 0 ),
    m_SectorIndex ( -1 ),
    m_RecordPtr ( NULL )
{
    FUNCTION_ENTRY ( this, "cDiskFile ctor", true );

    m_TotalRecordsLeft = ( m_FDR->FileStatus & PROGRAM_TYPE ) ? GetUSHORT ( &m_FDR->TotalSectors ) : GetUSHORT_LE ( &m_FDR->NoFixedRecords );

    memset ( m_SectorBuffer, 0, sizeof ( m_SectorBuffer ));
}

cDiskFile::~cDiskFile ()
{
    FUNCTION_ENTRY ( this, "cDiskFile dtor", true );
}

bool cDiskFile::ReadNextSector ()
{
    FUNCTION_ENTRY ( this, "cDiskFile::ReadNextSector", true );

    int totalSectors = GetUSHORT ( &m_FDR->TotalSectors );

    if ( m_SectorIndex >= totalSectors - 1 ) {
        return false;
    }

    ReadSector ( ++m_SectorIndex, m_SectorBuffer );

    m_RecordPtr   = m_SectorBuffer;
    m_RecordsLeft = m_FDR->RecordsPerSector;

    return true;
}

bool cDiskFile::SeekRecord ( int index )
{
    FUNCTION_ENTRY ( this, "cDiskFile::SeekRecord", true );

    if (( m_FDR->FileStatus & ( PROGRAM_TYPE | VARIABLE_TYPE )) != FIXED_TYPE ) {
        return false;
    }

    if ( index >= GetUSHORT_LE ( &m_FDR->NoFixedRecords )) {
        return false;
    }

    int sector = index / m_FDR->RecordsPerSector;
    ReadSector ( sector, m_SectorBuffer );

    int record = index - sector * m_FDR->RecordsPerSector;
    m_RecordPtr    = m_SectorBuffer + record * m_FDR->RecordLength;
    m_RecordsLeft -= record;

    return false;
}

int cDiskFile::ReadRecord ( void *buffer, int max )
{
    FUNCTION_ENTRY ( this, "cDiskFile::ReadRecord", true );

    if ( m_TotalRecordsLeft == 0 ) return -1;

    if ( m_RecordsLeft == 0 ) {
        ReadNextSector ();
    }

    int count = 0;

    if (( m_FDR->FileStatus & PROGRAM_TYPE ) == DATA_TYPE ) {

        int length = 0;

        if (( m_FDR->FileStatus & VARIABLE_TYPE ) == FIXED_TYPE ) {

            length = m_FDR->RecordLength;
            m_TotalRecordsLeft--;
            m_RecordsLeft--;

        } else {

            length = *m_RecordPtr++;
            if ( m_RecordPtr [length] == 0xFF ) {
                m_TotalRecordsLeft--;
                m_RecordsLeft = 0;
            }

        }

        count = ( length < max ) ? length : max - 1;
        memcpy ( buffer, m_RecordPtr, count );
        m_RecordPtr += length;

        char *end = ( char * ) buffer + count;
        *end = '\0';

    } else {
        count = (( m_TotalRecordsLeft == 1 ) && ( m_FDR->EOF_Offset != 0 )) ? m_FDR->EOF_Offset : DEFAULT_SECTOR_SIZE;
        memcpy ( buffer, m_RecordPtr, count );
        m_TotalRecordsLeft--;
        m_RecordsLeft = 0;
    }

    return count;
}

int cDiskFile::WriteRecord ( void *, int )
{
    FUNCTION_ENTRY ( this, "cDiskFile::WriteRecord", true );

    // TBD

    return 0;
}

int cDiskFile::ReadSector ( int index, void *buffer )
{
    FUNCTION_ENTRY ( this, "cDiskFile::ReadSector", true );

    int totalSectors = GetUSHORT ( &m_FDR->TotalSectors );
    if ( index >= totalSectors ) return -1;

    const sSector *sector = m_Disk->GetFileSector ( m_FDR, index );
    memcpy ( buffer, sector->Data, DEFAULT_SECTOR_SIZE );
    return 0;
}

int cDiskFile::WriteSector ( int index, void *buffer )
{
    FUNCTION_ENTRY ( this, "cDiskFile::WriteSector", true );

    int totalSectors = GetUSHORT ( &m_FDR->TotalSectors );
    if ( index >= totalSectors ) {
        int count = index - totalSectors + 1;
        if ( m_Disk->ExtendFile ( m_FDR, count ) != count ) return -1;
    }

    sSector *sector = m_Disk->GetFileSector ( m_FDR, index );
    if ( memcmp ( sector->Data, buffer, DEFAULT_SECTOR_SIZE ) != 0 ) {
        memcpy ( sector->Data, buffer, DEFAULT_SECTOR_SIZE );
        m_Disk->DiskModified ();
    }
    
    return 0;
}


/*
class cFile {

    bool      m_IsNative;
    bool      m_IsVariable;
    int       m_RecordSize;
    int       m_RecordsPerSector;
    int       m_RecordsLeft;
    int       m_TotalRecordsLeft;
    ULONG     m_NextSector;
    FILE     *m_File;

public:

    cFile ( const char * );
    ~cFile ();

    int ReadRecord ( UCHAR *buffer, int size );
    
};

int cFile::ReadRecord ( UCHAR *buffer, int max )
{
    FUNCTION_ENTRY ( NULL, "cFile::ReadRecord", true );

    if ( m_TotalRecordsLeft-- == 0 ) return false;

    if ( m_RecordsLeft-- == 0 ) {
        m_RecordsLeft = m_RecordsPerSector - 1;
        fseek ( m_File, m_NextSector, SEEK_SET );
        m_NextSector += DEFAULT_SECTOR_SIZE;
    }

    char temp [DEFAULT_SECTOR_SIZE];

    if ( m_IsNative == true ) {
    } else {

        int length = m_RecordSize;
        if ( m_IsVariable == true ) length = fgetc ( m_File );

        if ( length < max ) {
            buffer [length] = '\0';
        } else {
            fread ( temp, length, 1, m_File );
            memcpy ( buffer, temp, max - 1 );
            buffer [max-1] = '\0';
        }
    }

    return true;
}
*/
/*
    sFileDescriptorRecord fdr;
    memset ( &fdr, 0, sizeof ( fdr ));
    fread ( &fdr, 128, 1, file );

    if (( fdr.FileStatus & ( VARIABLE_TYPE | INTERNAL_TYPE | PROGRAM_TYPE )) != 0 ) {
        fprintf ( stderr, "PROGRAM/INTERNAL/VARIABLE file types are not supported\n" );
        return false;
    }

    // We'll use this to keep track of the # of records left
    sRecordInfo info;
    info.IsVariable       = ( fdr.FileStatus & VARIABLE_TYPE ) ? true : false;
    info.RecordSize       = fdr.RecordLength;
    info.RecordsPerSector = fdr.RecordsPerSector;
    info.RecordsLeft      = fdr.RecordsPerSector;
    info.TotalRecordsLeft = fdr.NoFixedRecords;
    info.NextSector       = ftell ( file ) + DEFAULT_SECTOR_SIZE;
*/
